@@ -1,7 +1,10 @@
 import requests
+import json
+
 
 class GAuthifyError(Exception):
     """All Errors"""
+
     def __init__(self, msg, http_status, error_code, response_body):
         super(GAuthifyError, self).__init__(msg)
         self.msg = msg
@@ -30,6 +33,7 @@ class NotFoundError(GAuthifyError):
     """
     pass
 
+
 class ServerError(GAuthifyError):
     """
     Raised for connection issues and any other error that the server can
@@ -42,6 +46,7 @@ class RateLimitError(GAuthifyError):
     """
     Raised when API limit reached either by lack of payment or membership limit
     """
+    pass
 
 
 class GAuthify(object):
@@ -59,43 +64,76 @@ class GAuthify(object):
         """
         Handles API Requests
         """
+        json_resp, status_code, req = {}, '', ''
         for base_url in self.access_points:
             try:
                 req_url = base_url + url_addon
                 req = requests.request(type.lower(), req_url, data=params,
-                               params=params, headers=self.headers, timeout=1.5)
+                                       params=params, headers=self.headers,
+                                       timeout=1.5)
                 status_code = req.status_code
                 json_resp = req.json
                 if not isinstance(json_resp, dict) or (status_code > 400 and
-                               status_code not in [401, 402, 406, 404]):
+                                status_code not in [401, 402, 406, 404]):
                     raise requests.ConnectionError
                 break
             except requests.RequestException, e:
                 if base_url == self.access_points[-1]:
                     raise ServerError("Communication error with all access"
-                                      "points. Please contact support@gauthify.com for help.", 500, '500', req.raw)
+                                      "points. Please contact "
+                                      "support@gauthify.com for help.",
+                                      500, '500', '')
                 continue
         if status_code == 401:
-            raise ApiKeyError(json_resp['error_message'], status_code, json_resp['error_code'], req.raw)
+            raise ApiKeyError(json_resp['error_message'], status_code,
+                              json_resp['error_code'], req.raw)
         if status_code == 402:
-            raise RateLimitError(json_resp['error_message'], status_code, json_resp['error_code'], req.raw)
+            raise RateLimitError(json_resp['error_message'], status_code,
+                                 json_resp['error_code'], req.raw)
         elif status_code == 406:
-            raise ParameterError(json_resp['error_message'], status_code, json_resp['error_code'], req.raw)
+            raise ParameterError(json_resp['error_message'], status_code,
+                                 json_resp['error_code'], req.raw)
         elif status_code == 404:
-            raise NotFoundError(json_resp['error_message'], status_code, json_resp['error_code'], req.raw)
+            raise NotFoundError(json_resp['error_message'], status_code,
+                                json_resp['error_code'], req.raw)
         return json_resp.get('data')
 
 
-
-    def create_user(self, unique_id, display_name):
+    def create_user(self, unique_id, display_name, email=None,
+                    phone_number=None, meta=None):
         """
-        Creates or upserts a new user with a new secret key
+        Creates new user with a new secret key or resets if already exists
         """
         params = {'display_name': display_name}
+        if email:
+            params['email'] = email
+        if phone_number:
+            params['phone_number'] = phone_number
+        if meta:
+            params['meta'] = json.dumps(meta)
         url_addon = "users/{}/".format(unique_id)
         return self.request_handler(
             'post', url_addon=url_addon, params=params)
 
+
+    def update_user(self, unique_id, email=None, phone_number=None, meta=None,
+                    reset_key=False):
+        """
+        Updates user information
+        """
+        params = {}
+        if email:
+            params['email'] = email
+        if phone_number:
+            params['phone_number'] = phone_number
+        if meta:
+            params['meta'] = json.dumps(meta)
+        if reset_key:
+            params['reset_key'] = 'true'
+        url_addon = "users/{}/".format(unique_id)
+        return self.request_handler(
+            'put', url_addon=url_addon, params=params)
+
     def delete_user(self, unique_id):
         """
         Deletes user given by unique_id
@@ -113,7 +151,6 @@ class GAuthify(object):
         """
         Returns a single user, checks the auth_code if provided
         """
-
         url_addon = "users/{}/".format(unique_id)
         url_addon += 'check/{}/'.format(auth_code) if auth_code else ''
         return self.request_handler(
@@ -129,7 +166,7 @@ class GAuthify(object):
             if not response['provided_auth']:
                 raise ParameterError(
                     'auth_code not detected by server. Check if '
-                    'params sent via get request.')
+                    'params sent via get request.', '500', '', '')
             return response['authenticated']
         except GAuthifyError, e:
             if safe_mode:
@@ -138,24 +175,38 @@ class GAuthify(object):
                 raise e
 
 
-    def send_sms(self, unique_id, phone_number):
+    def get_user_by_token(self, token):
+        """
+        Returns a single user by ezAuth token
+        """
+        url_addon = "token/{}/".format(token)
+        return self.request_handler(
+            'get', url_addon=url_addon)
+
+
+    def send_sms(self, unique_id, phone_number=None):
         """
         Sends text message to phone number with the one time auth_code
         """
 
         url_addon = "users/{}/".format(unique_id)
-        url_addon += 'sms/{}/'.format(phone_number)
+        if phone_number:
+            url_addon += 'sms/{}/'.format(phone_number)
+        else:
+            url_addon += 'sms/'
         return self.request_handler(
             'get', url_addon=url_addon)
 
 
-    def send_email(self, unique_id, email):
+    def send_email(self, unique_id, email=None):
         """
         Sends email with the one time auth_code
         """
-
         url_addon = "users/{}/".format(unique_id)
-        url_addon += 'email/{}/'.format(email)
+        if email:
+            url_addon += 'email/{}/'.format(email)
+        else:
+            url_addon += 'email/'
         return self.request_handler(
             'get', url_addon=url_addon)
 
@@ -171,40 +222,46 @@ class GAuthify(object):
 
         print("1) Testing Creating a User...")
         result = self.create_user(account_name,
-                                  account_name)
-        print result
+                                  account_name, email='firsttest@gauthify.com',
+                                  phone_number='0123456789')
+        assert result['unique_id'] == account_name
+        assert result['display_name'] == account_name
+        assert result['email'] == 'firsttest@gauthify.com'
+        assert result['phone_number'] == '0123456789'
+        print(result)
         success()
 
         print("2) Retrieving Created User...")
         user = self.get_user(account_name)
         assert isinstance(user, dict)
-        print user
+        print(result)
         success()
 
         print("3) Retrieving All Users...")
         result = self.get_all_users()
         assert isinstance(result, list)
-        print result
+        print(result)
         success()
 
         print("4) Bad Auth Code...")
         result = self.check_auth(account_name, '112345')
         assert isinstance(result, bool)
-        print result
+        print(result)
         success()
 
         print("5) Testing one time pass (OTP)....")
         result = self.check_auth(account_name, user['otp'])
         assert isinstance(result, bool)
-        print result
+        print(result)
         if not result:
             raise ParameterError('Server error. OTP not working. Contact '
-                                 'support@gauthify.com for help.', 500, '500', '')
+                                 'support@gauthify.com for help.', 500, '500',
+                                 '')
         success()
         if test_email:
             print("5A) Testing email to {}".format(test_email))
             result = self.send_email(account_name, test_email)
-            print result
+            print(result)
             success()
         if test_number:
             print("5B) Testing SMS to {}".format(test_number))
@@ -213,19 +270,34 @@ class GAuthify(object):
         print("6) Detection of provided auth...")
         result = self.get_user(account_name, 'test12')
         assert result['provided_auth']
-        print result
+        print(result)
         success()
 
-        print("7) Deleting Created User...")
+        print("7) Testing updating email, phone, and meta")
+        result = self.update_user(account_name, email='test@gauthify.com',
+                                  phone_number='1234567890', meta={'a': 'b'})
+        assert result['email'] == 'test@gauthify.com'
+        assert result['phone_number'] == '1234567890'
+        assert result['meta']['a'] == 'b'
+        current_key = result['key']
+        success()
+
+        print("8) Testing key/secret")
+        result = self.update_user(account_name, reset_key=True)
+        print(current_key, result['key'])
+        assert result['key'] != current_key
+        success()
+
+        print("9) Deleting Created User...")
         result = self.delete_user(account_name)
         assert isinstance(result, bool)
-        print result
         success()
 
-        print("8) Testing backup server...")
+        print("10) Testing backup server...")
+        current = self.access_points[0]
         self.access_points[0] = 'https://blah.gauthify.com/v1/'
         result = self.get_all_users()
-        self.access_points[0] = 'https://api.gauthify.com/v1/'
+        self.access_points[0] = current
         print result
         success()
 
